import os
from pathlib import Path
import numpy as np
import yaml

from bbo.exceptions import NoDataException


def ndarray_representer(dumper: yaml.Dumper, array: np.ndarray) -> yaml.Node:
    return dumper.represent_list(array.tolist())


yaml.add_representer(np.ndarray, ndarray_representer)

version = 0.4


def update(labels, labeler="_unknown"):
    assert labels["version"] <= version, "Please update ACM traingui"

    # Before versioning
    if "version" not in labels or labels["version"] <= 0.2:
        if "labeler" not in labels:
            labels["labeler"] = {}
        if "labeler_list" not in labels:
            labels["labeler_list"] = [labeler]
        if labeler not in labels["labeler_list"]:
            labels["labeler_list"].append(labeler)

        labeled_frame_idxs = get_labeled_frame_idxs(labels)
        labeler_idx = labels["labeler_list"].index(labeler)

        for f_idx in labeled_frame_idxs:
            if f_idx not in labels["labeler"]:
                labels["labeler"][f_idx] = labeler_idx
            if f_idx not in labels["fr_times"]:
                labels["fr_times"][f_idx] = 0

    if labels["version"] <= 0.3:
        labeler = {}
        point_times = {}
        for ln in labels["labels"]:
            labeler[ln] = {}
            point_times[ln] = {}
            for fr_idx in labels['labels'][ln]:
                data_shape = get_data_shape(labels)
                labeler[ln][fr_idx] = np.ones(data_shape[0], dtype=np.uint16) * labels['labeler'][fr_idx]
                nanmask = np.any(np.isnan(labels["labels"][ln][fr_idx]), axis=1)
                labeler[ln][fr_idx][nanmask] = 0
                point_times[ln][fr_idx] = np.ones(data_shape[0], dtype=np.uint64) * labels['fr_times'][fr_idx]

        labels["point_times"] = point_times
        labels["labeler"] = labeler
        labels.pop("fr_times")

    # Bring labeler list in shape (add specials etc)
    make_global_labeler_list([labels])

    labels["version"] = version
    return labels


def load(file_path):
    if isinstance(file_path, str):
        file_path = Path(file_path)

    if file_path.with_suffix(".yml").is_file():
        file_path = file_path.with_suffix(".yml")
        labels = load_raw_yaml(file_path)
    elif file_path.with_suffix(".npz").is_file():
        file_path = file_path.with_suffix(".npz")
        labels = np.load(file_path, allow_pickle=True)["arr_0"][()]
        print("WARNING: Loaded deprecated npz file!")
    else:
        raise FileNotFoundError(file_path.as_posix())

    labels = update(labels, labeler=file_path.parent.parent.stem)

    return labels


def load_raw_yaml(file_path: Path):
    with open(file_path.as_posix(), 'r') as f:
        labels = yaml.safe_load(f)

    for ln in labels["labels"]:
        for fr_idx in labels["labels"][ln]:
            labels["labels"][ln][fr_idx] = np.array(labels["labels"][ln][fr_idx])

    for ln in labels["labeler"]:
        for fr_idx in labels["labeler"][ln]:
            labels["labeler"][ln][fr_idx] = np.array(labels["labeler"][ln][fr_idx])

    for ln in labels["point_times"]:
        for fr_idx in labels["point_times"][ln]:
            labels["point_times"][ln][fr_idx] = np.array(labels["point_times"][ln][fr_idx])

    return labels


def save(file_path, labels, yml_only=False):
    if isinstance(file_path, str):
        file_path = Path(file_path)

    os.makedirs(file_path.parent.as_posix(), exist_ok=True)

    if not yml_only:
        np.savez(file_path.with_suffix(".npz").as_posix(), labels)

    with open(file_path.with_suffix(".yml").as_posix(), 'w') as f:
        yaml.dump(labels, f, default_flow_style=None)


def get_labels(labels):
    return [lm for lm in labels["labels"] if len(labels["labels"][lm].keys()) > 0]


def get_labeled_frame_idxs(labels):
    frames = set()
    for ln in labels["labels"]:
        frames.update(set(labels["labels"][ln].keys()))

    return np.array(sorted(list(frames))).astype(int)


def get_labels_from_frame(labels, frame_idx):
    frame_labels = {}

    for lb in labels["labels"]:
        if frame_idx in labels["labels"][lb] and not np.all(np.isnan(labels["labels"][lb][frame_idx])):
            frame_labels[lb] = labels["labels"][lb][frame_idx]

    return frame_labels


def get_frame_labelers(labels, fr_idx, cam_idx=None):
    labeler_idxs = set()
    for ln in labels["labeler"]:
        if fr_idx in labels["labeler"][ln]:
            if cam_idx is None:
                labeler_idxs.update(labels["labeler"][ln][fr_idx])
            else:
                labeler_idxs.update(labels["labeler"][ln][fr_idx][(cam_idx,),])

    labelers = [labels["labeler_list"][i] for i in labeler_idxs]
    if "_unmarked" in labelers:
        labelers.pop(labelers.index("_unmarked"))
    return labelers


def merge(labels_list: list, target_file=None, overwrite=False, yml_only=False):
    # Load data from files
    labels_files = None
    if isinstance(labels_list[0], str):
        labels_files = [Path(lf).expanduser().resolve() for lf in labels_list]
    elif isinstance(labels_list[0], Path):
        labels_files = labels_list
    else:
        assert target_file is not None, "target_file is only supported if labels_list contains paths"

    # Normalize path of target_file
    if isinstance(target_file, str):
        target_file = Path(target_file).expanduser().resolve()

    # Add target files as first labels file if existing
    if target_file is not None and target_file.is_file():
        labels_files.insert(0, target_file)

    # Load data from labels files
    if labels_files is not None:
        labels_list = [load(lf.as_posix()) for lf in labels_files]

    make_global_labeler_list(labels_list)

    # Merge file-wise
    target_labels = labels_list[0]
    data_shape = get_data_shape(target_labels)

    for labels in labels_list[1:]:
        initialize_target(labels, target_labels, data_shape)

        for ln in labels["labels"]:
            for fr_idx in labels["labels"][ln]:
                target_cam_mask = target_labels["labels"][ln][fr_idx] != 0
                source_cam_mask = labels["labels"][ln][fr_idx] != 0
                source_newer_mask = target_labels["point_times"][ln][fr_idx] < labels["point_times"][ln][fr_idx]

                replace_mask = source_cam_mask & source_newer_mask

                if not overwrite:
                    replace_mask &= (~target_cam_mask)

                target_labels["labels"][ln][fr_idx][replace_mask] = labels["labels"][ln][fr_idx][replace_mask]
                target_labels["labeler"][ln][fr_idx][replace_mask] = labels["labeler"][ln][fr_idx][replace_mask]
                target_labels["point_times"][ln][fr_idx][replace_mask] = labels["point_times"][ln][fr_idx][replace_mask]

    sort_dictionaries(target_labels)

    if target_file is not None:
        save(target_file, target_labels, yml_only=yml_only)
    return target_labels


def combine_cams(labels_list: list, target_file=None, yml_only=False):
    # Normalize path of target_file
    if isinstance(target_file, str):
        target_file = Path(target_file).expanduser().resolve()

    # Load data from files
    for i_l, label in enumerate(labels_list):
        if isinstance(label, str):
            if label == "None":
                labels_list[i_l] = None
                continue
            labels_list[i_l] = Path(label)
        if isinstance(label, Path):
            labels_list[i_l] = labels_list[i_l].expanduser().resolve()
        labels_list[i_l] = load(labels_list[i_l].as_posix())

    make_global_labeler_list(labels_list)

    target_labels = get_empty_labels()
    target_labels['labeler_list'] = make_global_labeler_list(labels_list)

    data_shape = (len(labels_list), 2)

    for cam_idx, labels in enumerate(labels_list):
        if labels is None:
            continue

        # Walk through frames
        initialize_target(labels, target_labels, data_shape)

        for ln in labels["labels"]:
            for fr_idx in labels["labels"][ln]:
                target_labels["labels"][ln][fr_idx][cam_idx] = labels["labels"][ln][fr_idx]
                target_labels["point_times"][ln][fr_idx][cam_idx] = labels["point_times"][ln][fr_idx]
                target_labels["labeler"][ln][fr_idx][cam_idx] = labels["labeler"][ln][fr_idx]

    sort_dictionaries(target_labels)

    if target_file is not None:
        save(target_file, target_labels, yml_only=yml_only)
    return target_labels


def initialize_target(labels, target_labels, data_shape):
    # Walk through frames
    for ln in labels["labels"]:
        # Initialize label key
        if ln not in target_labels["labels"]:
            target_labels["labels"][ln] = {}
        if ln not in target_labels["labeler"]:
            target_labels["labeler"][ln] = {}
        if ln not in target_labels["point_times"]:
            target_labels["point_times"][ln] = {}
        for fr_idx in labels["labels"][ln]:
            # Initialize frame index
            if fr_idx not in target_labels["labels"][ln]:
                target_labels["labels"][ln][fr_idx] = \
                    np.full(data_shape, np.nan)
            if fr_idx not in target_labels["labeler"][ln]:
                target_labels["labeler"][ln][fr_idx] = \
                    np.full(data_shape[0], 0, dtype=np.uint16)
            if fr_idx not in target_labels["point_times"][ln]:
                target_labels["point_times"][ln][fr_idx] = \
                    np.full(data_shape[0], 0, dtype=np.uint64)


def sort_dictionaries(target_labels):
    # Sort dictionaries
    for label in target_labels["labeler"]:
        target_labels["labeler"][label] = dict(sorted(target_labels["labeler"][label].items()))
    target_labels["labeler"] = dict(sorted(target_labels["labeler"].items()))

    for label in target_labels["point_times"]:
        target_labels["point_times"][label] = dict(sorted(target_labels["point_times"][label].items()))
    target_labels["point_times"] = dict(sorted(target_labels["point_times"].items()))

    for label in target_labels["labels"]:
        target_labels["labels"][label] = dict(sorted(target_labels["labels"][label].items()))
    target_labels["labels"] = dict(sorted(target_labels["labels"].items()))


def make_global_labeler_list(labels_list):
    # This changes labeler_list in place!!!
    # Create a new global list of all labelers
    labeler_list_all = []
    for labels in labels_list:
        if labels is None:
            continue
        if "labeler_list" in labels:
            labeler_list_all += labels["labeler_list"]
    labeler_list_all += ["_unknown"]
    labeler_list_all += ["_unmarked"]

    # Make unique and sorted
    labeler_list_all = sorted(list(set(labeler_list_all)))
    # Get specials to the front
    labeler_list_all.pop(labeler_list_all.index("_unmarked"))
    labeler_list_all.pop(labeler_list_all.index("_unknown"))
    labeler_list_all.insert(0, "_unknown")
    labeler_list_all.insert(0, "_unmarked")

    # Rewrite to global index list
    for labels in labels_list:
        if labels is None:
            continue
        for ln in labels["labels"]:
            for fr_idx in labels['labels'][ln]:
                for i, labeler_idx in enumerate(labels['labeler'][ln][fr_idx]):
                    labeler = labels["labeler_list"][labeler_idx]
                    labels['labeler'][ln][fr_idx][i] = labeler_list_all.index(labeler)
        labels["labeler_list"] = labeler_list_all
    return labeler_list_all


def get_data_shape(labels):
    for ln in labels['labels']:
        for fr_idx in labels['labels'][ln]:
            return labels['labels'][ln][fr_idx].shape
    raise NoDataException("Did not find any data to determine data shape")


def get_n_cams(labels):
    return get_data_shape(labels)[0]


def get_empty_labels():
    return {
        'labels': {},
        'point_times': {},
        'labeler_list': ["_unmarked", "_unknown"],
        'labeler': {},
        'version': version,
    }


def to_array(labels,
             extract_frame_idxs=None, extract_labels=None,  # Extract only these parts of data
             time_bases=None,  # Rearrange after these time bases for cams
             label_identity=None  # Treat as identical labels (nanmean if both are labeled)
             ):
    cams_n = get_n_cams(labels)

    if extract_frame_idxs is None:
        extract_frame_idxs = get_labeled_frame_idxs(labels)

    if extract_labels is None:
        extract_labels = get_labels(labels)
    time_base = np.unique(np.concatenate(time_bases, axis=0))

    if time_bases is None:
        time_bases = [np.arange(len(extract_frame_idxs)) for _ in range(cams_n)]

    # cams x frames x landmarks x pix_coords
    landmark_imcoords = np.full((cams_n, len(time_base), len(extract_labels), 2), np.nan)
    for i_lm, lm in enumerate(extract_labels):
        if lm in labels["labels"]:
            for i_fr, fr_idx in enumerate(extract_frame_idxs):
                # Map identities
                if label_identity is not None and label_identity[i_lm] is not None:
                    lm_idx = extract_labels.index(label_identity[i_lm])
                else:
                    lm_idx = i_lm

                for i_cam in range(cams_n):
                    cam_time = time_bases[i_cam][i_fr]
                    time_base_idx = np.where(time_base == cam_time)[0][0]
                    if fr_idx in labels["labels"][lm]:
                        cam_coords = labels["labels"][lm][fr_idx][i_cam]
                        if not np.any(np.isnan(cam_coords)):
                            # if (fr_idx == 7529 and i_cam == 1) or (fr_idx == 7572 and i_cam == 0):
                            #     print(f"Writing cam {i_cam}, frame {fr_idx} to {time_base_idx}")
                            landmark_imcoords[i_cam, time_base_idx, lm_idx] = (
                                np.nanmean(np.array(
                                    [cam_coords, landmark_imcoords[i_cam, time_base_idx, lm_idx]]
                                ), axis=0))
    return landmark_imcoords, time_base





import yaml
from pathlib import Path
import re
import os
from pprint import pprint
import sys
from bbo.path_management import get_replace_dict

uuid = "f1f5ba7b_baf1_4f83_b940_a5f9e8468c45"


def load(yml_path, replace_dict=None, replace_dict_set_file=False, include=True, dependencies=None, ignore_missing=False):
    if replace_dict is None:
        replace_dict = get_replace_dict()

    if isinstance(yml_path, str):
        yml_path = Path(yml_path)

    yml_path.expanduser().resolve()

    if not yml_path.is_file():
        raise FileNotFoundError(yml_path.as_posix())

    try:
        with open(yml_path, 'r') as file:
            yaml_string = file.read()
    except:
        print(yml_path)
        raise

    # !include cannot be read by the yaml reader, we replace with a unique id
    yaml_string = re.sub(r'(\s*)!include:', r"\1" + uuid + ":", yaml_string)
    yaml_dict = yaml.safe_load(yaml_string)
    if yaml_dict is None:
        yaml_dict = {}

    # TODO: It might be worth checking https://stackoverflow.com/questions/528281/how-can-i-include-a-yaml-file-inside-another
    replace_include(yaml_dict, yml_path.parent, replace_dict, include, dependencies, ignore_missing)

    if "file" not in replace_dict or replace_dict_set_file:
        replace_dict["file"] = yml_path.parent.as_posix()
    yaml_dict = replace_placeholders(yaml_dict, replace_dict)

    return yaml_dict


def replace_include(yaml_dict: dict, path: Path, replace_dict=None,
                    include=True, dependencies=None, ignore_missing=False):
    for key in list(yaml_dict.keys()):  # Keys will change during iteration
        if key == uuid:
            subdict_path = Path(yaml_dict[key]).expanduser()

            if not subdict_path.is_absolute():
                subdict_path = (path / subdict_path).expanduser().resolve()
            else:
                subdict_path = subdict_path.resolve()
            if dependencies is not None:
                dependencies.add(subdict_path)
            if include and (not ignore_missing or subdict_path.is_file()):
                yaml_dict.pop(uuid)
                subdict = load(subdict_path.as_posix(), replace_dict, replace_dict_set_file=True,
                               include=include, dependencies=dependencies, ignore_missing=ignore_missing)

                for subkey in subdict:
                    if subkey in yaml_dict:
                        merge_dicts(yaml_dict[subkey], subdict[subkey])
                    else:
                        yaml_dict[subkey] = subdict[subkey]
        elif isinstance(yaml_dict[key], dict):
            replace_include(yaml_dict[key], path, replace_dict, include, dependencies, ignore_missing)

    return yaml_dict


def merge_dicts(target_dict, merge_dict):
    for key in merge_dict:
        if key in target_dict:
            if isinstance(target_dict[key], dict) and isinstance(merge_dict[key], dict):
                merge_dicts(target_dict[key], merge_dict[key])
            else:
                target_dict[key] = merge_dict[key]
        else:
            target_dict[key] = merge_dict[key]


def replace_placehodlers_recursive(entry, replace_list):
    if isinstance(entry, str):
        for rep in replace_list:
            entry = entry.replace(rep[0], rep[1])
    if isinstance(entry, dict):
        for key in entry:
            entry[key] = replace_placehodlers_recursive(entry[key], replace_list)
    if isinstance(entry, list):
        for i in range(len(entry)):
            entry[i] = replace_placehodlers_recursive(entry[i], replace_list)
    return entry


def replace_placeholders(yaml_dict: dict, replace_dict=None):
    if replace_dict == None or len(replace_dict) == 0:
        return yaml_dict
    replace_list = [("{"+item[0]+"}",item[1]) for item in replace_dict.items()]
    return replace_placehodlers_recursive(yaml_dict, replace_list)



from pathlib import Path
import yaml
import sys


def get_custom_replace_dict():
    config_file = Path("~/.bbo/replace_dict.yml").expanduser().resolve()

    # Create config file if it doesn't exist
    if not config_file.is_file():
        with open(config_file, "w"):
            pass

    # Open config file
    with open(config_file, "r") as yaml_file:
        custom_replace_dict = yaml.safe_load(yaml_file)
    # An empty file returns None, fix that
    if custom_replace_dict is None:
        custom_replace_dict = {}
    # Python does not know ~, resolve that
    for key in custom_replace_dict:
        custom_replace_dict[key] = Path(custom_replace_dict[key]).expanduser().resolve().as_posix()

    return custom_replace_dict


def get_default_replace_dict():
    default_replace_dict = {
        'dropbox': Path("~/Dropbox/").expanduser().resolve().as_posix()
    }

    if sys.platform.startswith('linux'):
        for path in [
            "/media/smb/soma-fs.ad01.caesar.de/bbo/",
            "/media/smb/soma.ad01.caesar.de/bbo/"
        ]:
            try:
                if Path(path).is_dir():
                    default_replace_dict["storage"] = path
                    break
            except OSError as e:
                if e.errno == 116:  # Stale file handle, sometimes happens with SOMA
                    replace_dict["storage"] = path
    elif sys.platform.startswith('win') or sys.platform.startswith('cygwin'):
        default_replace_dict["storage"] = "s:/"
    else:
        print("Warning unknown platform", sys.platform, file=sys.stderr)

    return default_replace_dict


def get_replace_dict():
    return get_default_replace_dict() | get_custom_replace_dict()



import pathlib
from setuptools import find_packages, setup
# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="bbo_bbo",
    version="0.1.1",
    description="Base tools for MPINB BBO group",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/bbo-lab/bbo",
    author="BBO-lab @ caesar",
    author_email="kay-michael.voit@caesar.de",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=['bbo'],
    include_package_data=True,
    install_requires=["pyyaml"],
)
